==> intel-graphics-compiler_1.0.3627.orig.tar.gz: intel-graphics-compiler-igc-1.0.3627/visa/iga/IGALibrary/Frontend/Formatter.hpp <==
/*===================== begin_copyright_notice ==================================

Copyright (c) 2017 Intel Corporation

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


======================= end_copyright_notice ==================================*/
#ifndef _IGA_FORMATTER
#define _IGA_FORMATTER

#include "Floats.hpp"
#include "../ErrorHandler.hpp"
#include "../IR/DUAnalysis.hpp"
#include "../IR/Kernel.hpp"
#include "../strings.hpp"

#include <iomanip>
#include <ostream>
#include <string>
#include <sstream>

namespace iga {
    //////////////////////////////////////////////////////////////////////////
    // EXTERNAL INTERFACES
    //////////////////////////////////////////////////////////////////////////

    typedef const char  *(*LabelerFunction)(int32_t, void *);

    // Formatter options
    struct FormatOpts {
        Platform          platform;
        LabelerFunction   labeler = nullptr;
        void             *labelerContext = nullptr;
        bool              numericLabels = false;
        bool              hexFloats = true;
        bool              printInstPc = false;
        bool              syntaxExtensions = false;
        // swsb encoding mode, if not specified, the encoding mode will be
        // derived from platform by SWSB::getEncodeMode
        SWSB_ENCODE_MODE  swsbEncodingMode = SWSB_ENCODE_MODE::SWSBInvalidMode;
        bool              printInstDeps = true;
        bool              printInstBits = true;
        bool              printLdSt = false;
        DepAnalysis      *liveAnalysis = nullptr;

        // format with default labels
        FormatOpts(Platform _platform)
            : platform(_platform)
        {
        }

        // format with all parameters settable
        // requires at least a context
        FormatOpts(
            Platform _platform,
            LabelerFunction _labeler,
            void *_labelerCtx = nullptr,
            bool _numericLabels = false,
            bool _printInstPc = false,
            bool _hexFloats = true,
            bool _syntaxExtensions = false)
            : platform(_platform)
            , labeler(_labeler)
            , labelerContext(_labelerCtx)
            , numericLabels(_numericLabels)
            , hexFloats(_hexFloats)
            , printInstPc(_printInstPc)
            , syntaxExtensions(_syntaxExtensions)
        {
        }

        void setSWSBEncodingMode(SWSB_ENCODE_MODE mode)
        {
            swsbEncodingMode = mode;
        }
    };


    void FormatKernel(
        ErrorHandler &e,
        std::ostream &o,
        const FormatOpts &opts,
        const Kernel &k,
        const void *bits = nullptr);

    void FormatInstruction(
        ErrorHandler &e,
        std::ostream &o,
        const FormatOpts &opts,
        const Instruction &i,
        const void *bits = nullptr);


#ifndef DISABLE_ENCODER_EXCEPTIONS
    // this uses the decoder, which uses exceptions
    void FormatInstruction(
        ErrorHandler &e,
        std::ostream &o,
        const FormatOpts &opts,
        size_t startPc,
        const void *bits,
        bool useNativeDecoder = false);
    void FormatInstruction(
        ErrorHandler &e,
        std::ostream &o,
        const FormatOpts &opts,
        const void *bits);
#endif

    void GetDefaultLabelName(
        std::ostream &o,
        int32_t pc);

    /////////////////////////////////////////////////////////
    // Stuff mainly for testing and debugging.

    // for debugging, shows the opcode and bytes
    std::string FormatOpBits(const iga::Model &m, const void *bits);
    // decodes the op for this platform
    std::string FormatOpName(const iga::Model &m, const void *bits);






    //////////////////////////////////////////////////////////////////////////
    // Used by other modules internal to this directory
    //////////////////////////////////////////////////////////////////////////




    // The abstract implementation of a formatter that has a notion of
    // column alignment and some other basic, language-agnostic constructs.
    class BasicFormatter {
        size_t           currColLen; // length of the current column
        size_t           currColStart; // stream offset of start of last col
        size_t           currColDebt; // how far the current column has overflowed
    protected:
        std::ostream&    o;
        BasicFormatter(std::ostream &out) :
            currColLen(0),
            currColStart(0),
            currColDebt(0),
            o(out)
        {
        }
    public:

        // start or finish a padded column
        void startColumn(int len) {
            currColLen = len;
            currColStart = (size_t)o.tellp();
        }
        void finishColumn() {
            size_t end = (size_t)o.tellp();
            size_t actualWidth = end - currColStart;
            if (actualWidth <= currColLen) {
                // we underflowed; we have space left;
                // if there is debt, we can cover some of it
                size_t padding = currColLen - actualWidth;
                if (currColDebt > 0) {
                    size_t correction = std::min<size_t>(currColDebt, padding);
                    padding -= correction;
                    currColDebt -= correction;
                }
                emitSpaces(padding);
            } else if (actualWidth > currColLen) {
                // we overflowed accummulate debt
                currColDebt += actualWidth - currColLen;
            }
        }


        void newline() {
            emit('\n');
            currColDebt = 0;
        }

        template <typename T>
        void emit(const T &t) {
            o << t;
#ifdef TRACE_EMIT
            std::cerr << t;
            std::cerr.flush();
#endif
        }
        template <typename T, typename U>
        void emit(const T &t, const U &u) {
            emit(t); emit(u);
        }
        template <typename T, typename U, typename V>
        void emit(const T &t, const U &u, const V &v) {
            emit(t); emit(u); emit(v);
        }
        template <typename T, typename U, typename V, typename W>
        void emit(const T &t, const U &u, const V &v, const W &w) {
            emit(t); emit(u); emit(v); emit(w);
        }


        void emitSpaces(size_t n) {
            for (size_t i = 0; i < n; i++) {
                emit(' ');
            }
        }


        template <typename T>
        void emitDecimal(const T &t) {
            o << std::dec << t;
#ifdef TRACE_EMIT
            std::cerr << std::dec << t;
            std::cerr.flush();
#endif
        }


        template <typename T>
        void emitHex(const T &t, int w = 0) {
            fmtHex(o, (uint64_t)t);
#ifdef TRACE_EMIT
            std::cerr << "0x" << std::hex << t << std::dec;
            std::cerr.flush();
#endif
        }


        template <typename T>
        void emitBinary(const T &t) {
            emit("0b");
            std::stringstream ss;
            T bits = t;
            while (bits != 0) {
                emit((bits & 0x1) ? '1' : '0');
                bits >>= 1;
            }
            std::string str = ss.str();
            for (int i = (int)str.length() - 1; i >= 0; i--) {
                emit(str[i]);
            }
        }


        template <typename T>
        void emitFloat(const T &f) {
            FormatFloat(o, f);
        }
    };


    struct ColumnPreferences {
        int predication;
        int opCodeExecInfo;
        int flagMod;
        int dstOp;
        int srcOp;
        int sendDstOp; // we give send ops less space than normal
        int sendSrcOp; //we give send ops less space than norma
        int sendDesc;
        ColumnPreferences()
            : predication(8)
            // sends (16|M0)
            // madm (16|M16)
            , opCodeExecInfo(13)
            // (le)f0.0
            , flagMod(9)
            // r120.11<1>:hf
            , dstOp(12)
            , srcOp(16)
            , sendDstOp(7) // null:ud
                           // r120.4<4;4,1>:ub
            , sendSrcOp(6) // r120#1
            , sendDesc(10)
        {
        }
    };
} //end: namespace iga*

#endif

==> intel-graphics-compiler_1.0.5353.1.orig.tar.gz: intel-graphics-compiler-igc-1.0.5353.1/visa/iga/IGALibrary/Frontend/Formatter.hpp <==
/*===================== begin_copyright_notice ==================================

Copyright (c) 2017 Intel Corporation

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


======================= end_copyright_notice ==================================*/
#ifndef _IGA_FORMATTER
#define _IGA_FORMATTER

#include "Floats.hpp"
#include "../ErrorHandler.hpp"
#include "../IR/DUAnalysis.hpp"
#include "../IR/Kernel.hpp"
#include "../strings.hpp"

#include <iomanip>
#include <ostream>
#include <string>
#include <sstream>

namespace iga {
    //////////////////////////////////////////////////////////////////////////
    // EXTERNAL INTERFACES
    //////////////////////////////////////////////////////////////////////////

    typedef const char  *(*LabelerFunction)(int32_t, void *);

    // Formatter options
    struct FormatOpts {
        Platform          platform;
        LabelerFunction   labeler = nullptr;
        void             *labelerContext = nullptr;
        bool              numericLabels = false;
        bool              hexFloats = true;
        bool              printInstPc = false;
        bool              syntaxExtensions = false;
        // swsb encoding mode, if not specified, the encoding mode will be
        // derived from platform by SWSB::getEncodeMode
        SWSB_ENCODE_MODE  swsbEncodingMode = SWSB_ENCODE_MODE::SWSBInvalidMode;
        bool              printInstDeps = true;
        bool              printInstBits = true;
        bool              printLdSt = false;
        bool              printAnsi = false;
        DepAnalysis      *liveAnalysis = nullptr;

        // format with default labels
        FormatOpts(Platform _platform)
            : platform(_platform)
        {
        }

        // format with all parameters settable
        // requires at least a context
        FormatOpts(
            Platform _platform,
            LabelerFunction _labeler,
            void *_labelerCtx = nullptr,
            bool _numericLabels = false,
            bool _printInstPc = false,
            bool _hexFloats = true,
            bool _syntaxExtensions = false)
            : platform(_platform)
            , labeler(_labeler)
            , labelerContext(_labelerCtx)
            , numericLabels(_numericLabels)
            , hexFloats(_hexFloats)
            , printInstPc(_printInstPc)
            , syntaxExtensions(_syntaxExtensions)
        {
        }

        void setSWSBEncodingMode(SWSB_ENCODE_MODE mode) {
            swsbEncodingMode = mode;
        }

        // in iga.cpp
        void addApiOpts(uint32_t fmtOpts);
    };


    void FormatKernel(
        ErrorHandler &e,
        std::ostream &o,
        const FormatOpts &opts,
        const Kernel &k,
        const void *bits = nullptr);

    void FormatInstruction(
        ErrorHandler &e,
        std::ostream &o,
        const FormatOpts &opts,
        const Instruction &i,
        const void *bits = nullptr);


#ifndef IGA_DISABLE_ENCODER_EXCEPTIONS
    // this uses the decoder, which uses exceptions
    void FormatInstruction(
        ErrorHandler &e,
        std::ostream &o,
        const FormatOpts &opts,
        size_t startPc,
        const void *bits,
        bool useNativeDecoder = false);
    void FormatInstruction(
        ErrorHandler &e,
        std::ostream &o,
        const FormatOpts &opts,
        const void *bits);
#endif

    void GetDefaultLabelName(
        std::ostream &o,
        int32_t pc);

    /////////////////////////////////////////////////////////
    // Stuff mainly for testing and debugging.

    // for debugging, shows the opcode and bytes
    std::string FormatOpBits(const iga::Model &m, const void *bits);
    // decodes the op for this platform
    std::string FormatOpName(const iga::Model &m, const void *bits);


    //////////////////////////////////////////////////////////////////////////
    // Used by other modules internal to this directory
    //////////////////////////////////////////////////////////////////////////

    // This wraps a constant ansi escape sequence in a unique type so
    // template functions can handle them different from (const char *)'s.
    // We define emitT(ansi_esc) to not count the characters as part of the
    // current column (for column alignment).
    struct ansi_esc {
        const char *esc;
        constexpr ansi_esc() : ansi_esc(nullptr) {}
        constexpr ansi_esc(const char *_seq) : esc(_seq) {}
    };

    // This wraps a value of some sort to be emitted with an ANSI escape
    // sequence.  We emit this and an ANSI_RESET afterwards.
    // E.g. emit(ansi_span<int>("\033[1;32m",0x12))
    template <typename T>
    struct ansi_span {
        ansi_esc esc;
        const T &val;
        constexpr ansi_span(const T &_val)
            : ansi_span(nullptr, _val) { }
        constexpr ansi_span(const char *_seq, const T &_val)
            : esc(_seq), val(_val) { }
    };


    // The abstract implementation of a formatter that has a notion of
    // column alignment and some other basic, language-agnostic constructs.
    class BasicFormatter {
        size_t           currColCapacity; // preferred size of current column
        size_t           currColSize; // current col's size
        size_t           currLineDebt; // sum total of the column overflow
                                      // (later cols can pay for overflow
                                      // to realign future columns)
        bool             printAnsi;

    protected:
        ansi_esc ANSI_RESET;
        // subclasses to this class may define other ansi sequences

        std::ostream&    o;

        BasicFormatter(bool _printAnsi, std::ostream &out) :
            printAnsi(_printAnsi),
            currColCapacity((size_t)-1),
            currColSize(0),
            currLineDebt(0),
            o(out)
        {
            // TODO: could make these mappable via environment variable
            // export IGA_FormatAnsiRegisterArf="\033[38;2;138;43;211m"
            // export IGA_FormatAnsiMnemonic=...
            // export IGA_FormatAnsiComment=...
            if (printAnsi) {
                ANSI_RESET          = "\033[0m";
            }
        }

    public:
        // start or finish a padded column
        void startColumn(int len) {
            IGA_ASSERT(len >= 0, "negative column size");
            IGA_ASSERT(currColCapacity == (size_t)-1,
                "startColumn() called from with-in column "
                "(missing finishColumn())");
            currColCapacity = (size_t)len;
            currColSize = 0;
        }
        void finishColumn() {
            IGA_ASSERT(currColCapacity != (size_t)-1,
                "finishColumn() called with no active column "
                "(missing startColumn(...))");
            //
            if (currColSize <= currColCapacity) {
                // we underflowed; we have space left;
                // if there is debt, we can cover some of it
                size_t padding = currColCapacity - currColSize;
                if (currLineDebt > 0) {
                    size_t correction =
                        std::min<size_t>(currLineDebt, padding);
                    padding -= correction;
                    currLineDebt -= correction;
                }
                emitSpaces(padding);
            } else if (currColSize > currColCapacity) {
                // we overflowed accummulate debt
                currLineDebt += currColSize - currColCapacity;
            }
            //
            currColCapacity = (size_t)-1;
            currColSize = 0;
        }


        void newline() {
            IGA_ASSERT(currColCapacity == (size_t)-1,
                "column open at newline");
            emit('\n');
            currLineDebt = 0;
        }


        template <typename T>
        void emitT(ansi_span<T> t) {
            emitT(t.esc);
            emitT(t.value);
            emitT(ANSI_RESET);
        }
        void emitT(ansi_esc e) {
            // specialize this instance to not count the ANSI escapes as
            // output characters (size of column)
            if (e.esc)
                o << e.esc;
        }
        template <typename T>
        void emitT(const T &t) {
            size_t n = (size_t)o.tellp();
            o << t;
            currColSize += (size_t)o.tellp() - n;
        }


        // enables you emit a list of things
        //
        // emit(foo, bar, baz, ...);
        template <typename T, typename...Ts>
        void emit(T t) {
            emitT(t);
        }
        template <typename T, typename...Ts>
        void emit(T t, Ts...ts) {
            emitT(t);
            emit(ts...);
        }

        // emits an sequence conditionally surrounded by an ANSI color
        template <typename T, typename...Ts>
        void emitAnsi(bool z, ansi_esc esc, T t0, Ts...ts) {
            if (z)
                emitT(esc);

            emit(t0, ts...);

            if (z)
                emitT(ANSI_RESET);
        }
        template <typename T, typename...Ts>
        void emitAnsi(ansi_esc esc, T t, Ts...ts) {
            emitAnsi(true, esc, t, ts...);
        }

        void emitSpaces(size_t n) {
            for (size_t i = 0; i < n; i++)
                o << ' ';
            currColSize += n;
        }


        template <typename T>
        void emitDecimal(const T &t) {
            o << std::dec << t;
        }


        template <typename T>
        void emitHex(const T &t, int cw = 0) {
            fmtHex(o, (uint64_t)t, cw);
            o << std::dec;
        }


        template <typename T>
        void emitBinary(const T &t) {
            emit("0b");
            std::stringstream ss;
            T bits = t;
            while (bits != 0) {
                emit((bits & 0x1) ? '1' : '0');
                bits >>= 1;
            }
            std::string str = ss.str();
            for (int i = (int)str.length() - 1; i >= 0; i--) {
                emit(str[i]);
            }
        }


        template <typename T>
        void emitFloat(const T &f) {
            FormatFloat(o, f);
        }
    };


    struct ColumnPreferences {
        int predication;
        int opCodeExecInfo;
        int flagMod;
        int dstOp;
        int srcOp;
        int sendDstOp; // we give send ops less space than normal
        int sendSrcOp; //we give send ops less space than norma
        int sendDesc;
        ColumnPreferences()
            // enough space for (~f0.0) or (W), (W&~f0.0) overflows
            : predication(7)
            // sends (16|M0)
            // madm (16|M16)
            , opCodeExecInfo(13)
            // (le)f0.0
            , flagMod(9)
            // r120.11<1>:hf
            , dstOp(12)
            , srcOp(16)
            , sendDstOp(7) // null:ud
                           // r120.4<4;4,1>:ub
            , sendSrcOp(6) // r120#1
            , sendDesc(10)
        {
        }
    };
} //end: namespace iga*

#endif
